function BeneCast:OnInitialize()
	self:FillSpellInfo();
	
	-- Set BC_class to the player's class
	_, wv.PlayerClass = UnitClass('player');
	
	wv.DefaultConfig = self:GetDefaultConfig();
	self.db = LibStub('AceDB-3.0'):New('BeneCastDB',{ char = wv.DefaultConfig });
	LibStub('AceConfig-3.0'):RegisterOptionsTable('BeneCast', options, {'benecast', 'bc'});
			
	-- Reset the data if there BeneCast is severly outdated
	if table.getn(self.db.char.Profiles) == 0 or
	   not self.db.Setup.Version or
	   (self.db.Setup.Version < self.Constants.Version) then
		self:Print('Resetting settings');
		self.db.Char = _CopyTable(wv.DefaultConfig,true);
	end
	
	-- Load up spell data (and other data) once AddOn has been loaded
	self:ResetData();
	
end

function BeneCast:OnEnable()

	-- Register for Events
	self:RegisterEvent('ADDON_LOADED');
	self:RegisterEvent('PLAYER_ENTERING_WORLD');
	self:RegisterEvent('PLAYER_LEAVING_WORLD');
	self:RegisterEvent('VARIABLES_LOADED');
	self:RegisterEvent('SPELLS_CHANGED');
	self:RegisterEvent('PARTY_MEMBERS_CHANGED');
	self:RegisterEvent('PLAYER_TARGET_CHANGED');
	self:RegisterEvent('PLAYER_FOCUS_CHANGED');
	self:RegisterEvent('UNIT_AURA');
	self:RegisterEvent('PLAYER_AURAS_CHANGED');
	self:RegisterEvent('UNIT_AURASTATE');
	self:RegisterEvent('UPDATE_SHAPESHIFT_FORMS');
	self:RegisterEvent('PLAYER_ALIVE');
	self:RegisterEvent('PLAYER_DEAD');
	self:RegisterEvent('PLAYER_UNGHOST');
	self:RegisterEvent('UNIT_HEALTH');
	self:RegisterEvent('UNIT_PET');
	self:RegisterEvent('UNIT_TARGET');
	self:RegisterEvent('PLAYER_REGEN_DISABLED');
	self:RegisterEvent('PLAYER_REGEN_ENABLED');
		
end

function BeneCast:OnDisable()
	self:CancelAllTimers();
	self:UnRegisterAllEvents();
end

function BeneCast:GetProfile(pf, unit)
	if not BeneCastProfiles then
		-- Addon hasn't initialized yet
		return;
	end
	
	if not BeneCastProfiles[profileid] then
		self:ResetProfile(profileid);
	end
	
	if not BeneCastProfiles[profileid].SnapTo then
		BeneCastProfiles[profileid].SnapTo = {};
	end
	
	if unit then
		if not BeneCastProfiles[profileid].SnapTo[unit] then
			BeneCastProfiles[profileid].SnapTo[unit] = {
				Y = 0,
				X = 0,
				Point = 'TOPLEFT',
				RelativePoint = 'TOPLEFT',
				Orientation = BeneCastProfiles[profileid].Orientation,
			};
		end
		return BeneCastProfiles[profileid], BeneCastProfiles[profileid].SnapTo[unit];
	else
		return BeneCastProfiles[profileid];
	end
end

function BeneCast:GetCurrentProfileID()
	if wv.ConsiderInCombat then
		return 'combat';
	else
		return 'ooc';
	end
end

-- Function to resize the buttons of a profile
function BeneCast:ResizeButtonsForProfile(pf)

	if BeneCast.Panels[pf] == nil then
		BeneCast.Panels[pf] = {};
	end
	
	-- Resize the buttons of every panel
	for unit, panel in pairs (BeneCast.Panels[pf]) do
		panel:ResizeButtons(i);
	end
	
end

-- Resets data tables
function BeneCast:ResetData()

	-- Clear the BC_spell_data table
	_ClearTable(wv.SpellData);
	
	-- Reload the spell data
	self:LoadSpellData();
	
end

-- Function to make a shiny new, blank config
--  Doesn't mean that it will be used to actually reset the config though
function BeneCast:GetDefaultConfig()

	-- Reset the BeneCastProfiles-table
	-- assume this will be added to the char-part of an AceDB-database
	local newconfig = {
		Setup = {
			Debug = nil,
			Version = BeneCast.Constants.Version,
			SnapTo = 'BeneCast',
			HideMinimap = false,
			Unlock = false,
		},
		Profiles = {
			ooc = {
				name = 'Out of combat',
			},
			combat = {
				name = 'In combat',
			},
		},
	};
	
	self:ResetConfig(newconfig.Profiles.ooc);
	self:ResetConfig(newconfig.Profiles.combat);
	
	return newconfig;
	
end

function BeneCast:ResetConfig(profile)

	local profname = profile.name;
	_ClearTable(profile);
	if profname == nil then
		profname = '' .. GetTime();
	end 
	profile.name = profname;
	
	profile.SnapTo = {};
	profile.Config = {
		player = {},
		pet = {},
		party = {},
		partypet = {},
		other = {},
	};

	profile.ShowAllButtons = false;
	profile.FlashAsFade = false;
	profile.ShowTooltips = true;
	profile.ShowTooltipsName = false;
	profile.PlayerAsDefault = true;
	profile.ShowPets = false;
	profile.ShowRaid = false;
	profile.ShowRaidPets = false;
	profile.CheckRange = false; 
	profile.ShowTargetTarget = true;
	
	profile.ButtonSize = 36;

end

function BeneCast:Debug(msg)
	if ( BeneCastSetup.Debug and DEFAULT_CHAT_FRAME ) then
		self:Print('Debug: ' .. msg);
	end
end

function BeneCast:AddButtonThatFlashes(button)
	if not mv.ButtonsThatFlash[pf] then
		mv.ButtonsThatFlash[pf] = {};
	end
	if not mv.ButtonsThatFlash[pf][unit] then
		mv.ButtonsThatFlash[pf][unit] = {};
	end
	table.add(mv.ButtonsThatFlash[pf][unit],button);
	if wv.Timers.FlashButtons == nil then
		wv.Timers.FlashButtons = self:ScheduleRepeatingTimer(self.TimerFlashButtons,0.2,self);
	end
end

function BeneCast:UpdateAllPanels()

	for pfid, pfpanels in pairs(wv.Panels) do
		for unit, panel in pairs(pfpanels) do
			panel:Update();
		end
	end
	
end

function BeneCast:UpdatePanelsForUnit(unit)
	
	if unit then
		local panel;
		for pfid, pfpanels in pairs(wv.Panels) do
			panel = pfpanels[unit];
			if panel then
				panel:Update();
			end
		end
	end
	
end


-- *****************************************************************************
-- BeneCast Timers
-- *****************************************************************************

function BeneCast:TimerUpdatePanels()

	local todelete = {};
	for unit, when in pairs(wv.UnitsToUpdate) do
		if unit ~= 'count' and when ~= nil and (GetTime() > when) then
			-- can't remove items while traversing the table
			-- it would mess things up
			table.insert(todelete,unit);
			self:UpdatePanelsForUnit(unit);
		end
	end
	
	for k, unit in pairs(todelete) do
		wv.UnitsToUpdate[unit] = nil;
		wv.UnitsToUpdate.count = wv.UnitsToUpdate.count - 1;
	end
	if wv.UnitsToUpdate.count == 0 and wv.Timers.UpdatePanels ~= nil then
		self:CancelTimer(wv.Timers.UpdatePanels);
		wv.Timers.UpdatePanels = nil;
	end

end

function BeneCast:AddUnitToUpdate(unit, when)

	local utu = wv.UnitsToUpdate;
	if utu.count == 0 and wv.Timers.UpdatePanels == nil then
		wv.Timers.UpdatePanels = self:ScheduleRepeatingTimer(self.TimerUpdatePanels, 1, self);
	end
	if utu[unit] ~= nil then
		utu[unit] = when;
	else
		utu[unit] = when;
		utu.count = utu.count + 1;
	end

end

function BeneCast:TimerCheckEnchants()

	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges;
	hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo();
	
	-- Set the tooltip allright
	if not BeneCast_Tooltip:IsOwned(WorldFrame) then
		BeneCast_Tooltip:SetOwner(WorldFrame, 'ANCHOR_NONE');
	end
	
	local oldname, updatepanel = false;
	if ( hasMainHandEnchant ) then
		oldname = nil;
		if wv.MHWeaponEnchant then
			oldname = wv.MHWeaponEnchant.name;
		end
		wv.MHWeaponEnchant = self:CheckForEnchant('MainHandSlot');
		if wv.MHWeaponEnchant then
			wv.MHWeaponEnchant.duration = mainHandExpiration;
		end
		if wv.MHWeaponEnchant and wv.MHWeaponEnchant.name ~= oldname then
			updatepanel = true;
		end
	end
	if ( hasOffHandEnchant ) then
		oldname = nil;
		if wv.OHWeaponEnchant then
			oldname = wv.OHWeaponEnchant.name;
		end
		wv.OHWeaponEnchant = self:CheckForEnchant('SecondaryHandSlot');
		if wv.OHWeaponEnchant then
			wv.OHWeaponEnchant.duration = offHandExpiration;
		end
		if wv.OHWeaponEnchant and wv.OHWeaponEnchant.name ~= oldname then
			updatepanel = true;
		end
	end
	if updatepanel then
		self:UpdatePanelsForUnit('player');
	end

end

-- Function to return the spellinfo for the temp enchant on a weapon slot
function BeneCast:CheckForEnchant(slotname)

	local slotid, line, spellinfo;
	slotid = GetInventorySlotInfo(slotname);
	
	-- Set the tooltip to the slot
	BeneCast_Tooltip:ClearLines();
	BeneCast_Tooltip:SetInventoryItem('player', slotid);
	
	for i = 1, BeneCast_Tooltip:NumLines() do
		line = getglobal('BeneCast_TooltipTextLeft' .. i):GetText();
		if line then
			for name, group in pairs(wv.TempEnchantsInfo) do
				if string.find(name, line) then -- Like finding 'Flametongue' in 'Flametongue Weapon'
					spellinfo = wv.SpellInfo[group];
					if spellinfo then
						return _CopyTable(spellinfo,true);
					end
				end
			end
		end
	end
	
	return nil;
	
end

function BeneCast:TimerFlashButtons()

	-- Update all button-flashing at once
	local buffAlphaValue = nil;
	local buttonsintable = 0;
	for pfid, pfunits in pairs(wv.ButtonsThatFlash) do
		for unit, unitbuttons in pairs(pfunits) do
			for k, button in pairs(unitbuttons) do
				if buffAlphaValue == nil then --this way it only has to be calculated once
					if ( BuffFrameFlashState == 1 ) then
						buffAlphaValue = (BUFF_FLASH_TIME_ON - BuffFrameFlashTime) / BUFF_FLASH_TIME_ON;
						buffAlphaValue = buffAlphaValue * (1 - BUFF_MIN_ALPHA) + BUFF_MIN_ALPHA;
					else
						buffAlphaValue = BuffFrameFlashTime / BUFF_FLASH_TIME_ON;
						buffAlphaValue = (buffAlphaValue * (1 - BUFF_MIN_ALPHA)) + BUFF_MIN_ALPHA;
					end
				end
				button:SetAlpha(buffAlphaValue);
				buttonsintable = buttonsintable + 1;
			end
		end
	end
	if buttonsintable == 0 and wv.Timers.FlashButtons then
		self:CancelTimer(wv.Timers.FlashButtons);
		wv.Timers.FlashButtons = nil;
	end

end

function BeneCast:ToggleRangeTimer()

	if wv.Timers.CheckRange == nil then
		wv.Timers.CheckRange = self:ScheduleRepeatingTimer(self.TimerCheckRange, 0.5, self);
	else
		self:CancelTimer(wv.Timers.CheckRange);
		wv.Timers.CheckRange = nil;
	end

end

function BeneCast:TimerCheckRange()

	local inrange, pf, button, spellinfoid, range;
	pf = self:GetCurrentProfileID();
	for unit, panel in pairs(wv.Panels[pf]) do
		if unit ~= 'player' then
			if panel:IsVisible() and UnitExists(unit) then
				button = panel.next;
				while button do
					if button.spell then
						spellinfoid = button.spell.id;
						range = button.Range;
						inrange = IsSpellInRange(spellinfoid,unit);
						if (inrange == nil) or (inrange == 0) then
							range:Show();
							if unit == 'target' then
								PutDebugMsg(L['Spell %1 out of range for %2'](button.spell.name,unit));
							end
						else
							range:Hide();
						end
					end
					button = button.next;
				end
			end
		end
	end

end